from pathlib import Path
from typing import Dict, Any, Optional, List, Set, Union

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.types import Row
from pyspark.ml.param import Param
from pyspark.sql import Column
from pyspark.sql.dataframe import DataFrame

# noinspection PyUnresolvedReferences
from pyspark.sql.functions import col, to_json, from_json
from pyspark.sql.types import StructType
from pyspark.sql.functions import filter
from spark_fhir_schemas.r4.resources.healthcareservice import HealthcareServiceSchema
from spark_fhir_schemas.r4.resources.location import LocationSchema
from spark_fhir_schemas.r4.resources.organization import OrganizationSchema
from spark_fhir_schemas.r4.resources.practitioner import PractitionerSchema
from spark_fhir_schemas.r4.resources.practitionerrole import PractitionerRoleSchema
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FhirBundleSplitter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        temp_folder: Union[Path, str],
        view: Optional[str] = None,
        resource_types: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.resource_types: Param[Optional[List[str]]] = Param(
            self, "resource_types", ""
        )
        self._setDefault(resource_types=resource_types)

        self.temp_folder: Param[Union[Path, str]] = Param(self, "temp_folder", "")
        self._setDefault(temp_folder=temp_folder)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        resource_types: Optional[List[str]] = self.getResourceTypes()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        # temp_folder: Union[Path, str] = self.getTempFolder()

        if view:
            df = df.sql_ctx.table(view)

        # if caller did not pass resource types then get them from the data (albeit pretty slow)
        if resource_types is None:
            if progress_logger:
                progress_logger.write_to_log(
                    "WARNING: No resource types as passed so we will analyze the data to find resource types."
                    "This can be slow."
                )
            resource_types_rows: List[Row] = (
                df.select("entry.resource.resourceType").distinct().collect()
            )
            resource_types_set: Set[str] = set()
            resource_types_row: Row
            for resource_types_row in resource_types_rows:
                resource_types_list: List[str]
                for resource_types_list in resource_types_row:
                    resource_type: str
                    for resource_type in resource_types_list:
                        resource_types_set.add(resource_type)
            resource_types = list(resource_types_set)
            if progress_logger:
                progress_logger.write_to_log(
                    "Following resource types were found in the data:"
                )
                progress_logger.write_to_log(",".join(resource_types))

        # now select the columns
        column_specs: List[Column] = [
            from_json(
                to_json(  # take it through json to get the correct schema
                    filter(
                        "entry",
                        lambda x: x["resource"]["resourceType"] == resource_type,
                    ).getField("resource")
                ),
                schema=self.get_schema_for_resource(resource_type=resource_type),
            ).alias(resource_type.lower())
            for resource_type in resource_types
        ]

        all_columns: List[Column] = [col(c) for c in df.columns if c != "entry"]

        # add these columns
        df = df.select(*(all_columns + column_specs))

        if view:
            df.createOrReplaceTempView(view)

        return df

    @staticmethod
    def get_schema_for_resource(resource_type: str) -> StructType:
        # TODO: replace once we can have the FHIR schema project return schemas by name
        if resource_type == "Practitioner":
            schema = PractitionerSchema.get_schema()
            assert isinstance(schema, StructType), type(schema)
            return schema
        if resource_type == "PractitionerRole":
            schema = PractitionerRoleSchema.get_schema()
            assert isinstance(schema, StructType), type(schema)
            return schema
        if resource_type == "Organization":
            schema = OrganizationSchema.get_schema()
            assert isinstance(schema, StructType), type(schema)
            return schema
        if resource_type == "Location":
            schema = LocationSchema.get_schema()
            assert isinstance(schema, StructType), type(schema)
            return schema
        if resource_type == "HealthcareService":
            schema = HealthcareServiceSchema.get_schema()
            assert isinstance(schema, StructType), type(schema)
            return schema
        raise NotImplementedError(f"{resource_type}")

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResourceTypes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.resource_types)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTempFolder(self) -> Union[Path, str]:
        return self.getOrDefault(self.temp_folder)



from pyspark.ml import Transformer
from pyspark.sql import SparkSession

from library.features.carriers_python.v1.calculate import (
    FeatureCarrierPythonTransformer,
)
from library.features.carriers_python.v1.features_carriers_python_v1 import (
    FeaturesCarriersPythonV1,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_can_run_python_transformer(spark_session: SparkSession) -> None:
    # Arrange

    # Act
    with ProgressLogger() as progress_logger:
        result: Transformer = FeaturesCarriersPythonV1(
            parameters={"foo": "bar"}, progress_logger=progress_logger
        ).transformers[0]

    # Assert
    assert isinstance(result, FeatureCarrierPythonTransformer)

    # make sure we can call transform on it
    df = create_empty_dataframe(spark_session=spark_session)
    result.transform(df)



# noinspection PyProtectedMember
from pyspark.sql import SparkSession
from pyspark.sql.types import StringType, MapType


def register_udfs(session: SparkSession) -> None:
    """
    Adds registration for udfs
    :param session:
    """
    session.udf.registerJavaFunction(
        "runCqlLibrary",
        "com.bwell.services.spark.RunCqlLibrary",
        MapType(StringType(), StringType()),
    )



import shutil
from os import path
from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_json_exporter.v1.framework_json_exporter import (
    FrameworkJsonExporter,
)
from spark_pipeline_framework.transformers.framework_json_loader.v1.framework_json_loader import (
    FrameworkJsonLoader,
)
from tests.conftest import clean_spark_session


def test_can_export_simple_json_with_schema_on_delta(
    spark_session: SparkSession,
) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('schema_test.json')}"
    test_file_path_2: str = f"{data_dir.joinpath('schema_test_2.json')}"

    if path.isdir(data_dir.joinpath("temp")):
        shutil.rmtree(data_dir.joinpath("temp"))

    temp_folder = data_dir.joinpath("temp/")

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonLoader(view="books", file_path=test_file_path).transform(df)
    FrameworkJsonLoader(
        view="books_schema", file_path=test_file_path_2, use_schema_from_view="books"
    ).transform(df)
    result: DataFrame = spark_session.sql("SELECT * FROM books")
    result_2: DataFrame = spark_session.sql("SELECT * FROM books_schema")

    # Act
    assert result.schema == result_2.schema

    FrameworkJsonExporter(
        view="books_schema",
        file_path=temp_folder.joinpath("out.json"),
        delta_lake_table="table1",
    ).transform(df)

    df = spark_session.read.format("delta").load(str(temp_folder.joinpath("out.json")))
    assert df.count() == 2
    df.show(truncate=False)

    FrameworkJsonLoader(
        view="books_schema",
        file_path=temp_folder.joinpath("out.json"),
        delta_lake_table="table1",
    ).transform(df)

    result_2 = spark_session.sql("SELECT * FROM books_schema")
    assert result_2.count() == 2
    result_2.show(truncate=False)



import json
import math
from os import environ
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union, Callable

from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.rdd import RDD
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import Row
from pyspark.sql.utils import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_helpers import (
    send_fhir_delete,
    send_json_bundle_to_fhir,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_validation_exception import (
    FhirSenderValidationException,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


# noinspection PyProtectedMember


class FhirSender(FrameworkTransformer):
    FHIR_OPERATION_DELETE = "delete"
    FHIR_OPERATION_MERGE = "$merge"

    FHIR_OPERATIONS = [FHIR_OPERATION_DELETE, FHIR_OPERATION_MERGE]

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        server_url: str,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        response_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        resource: str,
        batch_size: Optional[int] = 30,
        limit: int = -1,
        validation_server_url: Optional[str] = None,
        throw_exception_on_validation_failure: Optional[bool] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        operation: str = FHIR_OPERATION_MERGE,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        error_view: Optional[str] = None,
        view: Optional[str] = None,
        ignore_status_codes: Optional[List[int]] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        num_partitions: Optional[int] = None,
        delta_lake_table: Optional[str] = None,
    ):
        """
        Sends FHIR json stored in a folder to a FHIR server

        :param server_url: url for the fhir server
        :param file_path: where to read the json from
        :param response_path: where to store the responses
        :param resource: name of resource to send to fhir server
        :param batch_size: how many resources to send in one batch call to fhir server
        :param validation_server_url: (Optional) url to use for validating FHIR resources
        :param throw_exception_on_validation_failure: whether to continue after a validation failure
        :param auth_server_url: url to auth server
        :param auth_client_id: client id
        :param auth_client_secret: client secret
        :param auth_login_token: login token to send auth server
        :param auth_scopes: scopes to request
        :param operation: What FHIR operation to perform (e.g. $merge, delete, etc.)
        :param mode: if output files exist, should we overwrite or append
        :param error_view: (Optional) log errors into this view (view only exists IF there are errors)
                            and don't throw exceptions.
                            schema: id, resourceType, issue
        :param view: (Optional) store merge result in this view
        :param delta_lake_table: use delta lake format
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)

        assert file_path

        assert progress_logger

        assert server_url

        self.logger = get_logger(__name__)

        self.server_url: Param[str] = Param(self, "server_url", "")
        self._setDefault(server_url=None)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.response_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "response_path", "")
        self._setDefault(response_path=None)

        self.resource: Param[str] = Param(self, "resource", "")
        self._setDefault(resource=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.throw_exception_on_validation_failure: Param[Optional[bool]] = Param(
            self, "throw_exception_on_validation_failure", ""
        )
        self._setDefault(
            throw_exception_on_validation_failure=throw_exception_on_validation_failure
        )

        self.validation_server_url: Param[Optional[str]] = Param(
            self, "validation_server_url", ""
        )
        self._setDefault(validation_server_url=None)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[List[str]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.operation: Param[str] = Param(self, "operation", "")
        self._setDefault(operation=operation)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.retry_count: Param[Optional[int]] = Param(self, "retry_count", "")
        self._setDefault(retry_count=None)

        self.exclude_status_codes_from_retry: Param[Optional[List[int]]] = Param(
            self, "exclude_status_codes_from_retry", ""
        )
        self._setDefault(exclude_status_codes_from_retry=None)

        self.num_partitions: Param[Optional[int]] = Param(self, "num_partitions", "")
        self._setDefault(num_partitions=None)

        self.ignore_status_codes: Param[Optional[List[int]]] = Param(
            self, "ignore_status_codes", ""
        )
        self._setDefault(ignore_status_codes=None)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        file_path: Union[
            Path, str, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getFilePath()
        if callable(file_path):
            file_path = file_path(self.loop_id)
        response_path: Union[
            Path, str, Callable[[Optional[str]], Union[Path, str]]
        ] = self.getResponsePath()
        if callable(response_path):
            response_path = response_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        resource_name: str = self.getResource()
        server_url: str = self.getServerUrl()
        batch_size: Optional[int] = self.getBatchSize()
        throw_exception_on_validation_failure: Optional[
            bool
        ] = self.getThrowExceptionOnValidationFailure()
        operation: str = self.getOperation()
        mode: str = self.getMode()

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        if not batch_size or batch_size == 0:
            batch_size = 30
        else:
            batch_size = int(batch_size)

        validation_server_url: Optional[str] = self.getValidationServerUrl()
        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()
        auth_access_token: Optional[str] = None

        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view: Optional[str] = self.getOrDefault(self.view)

        log_level: Optional[str] = environ.get("LOGLEVEL")

        ignore_status_codes: List[int] = self.getIgnoreStatusCodes() or []
        ignore_status_codes.append(200)

        retry_count: Optional[int] = self.getOrDefault(self.retry_count)
        exclude_status_codes_from_retry: Optional[List[int]] = self.getOrDefault(
            self.exclude_status_codes_from_retry
        )

        # num_partitions: Optional[int] = self.getOrDefault(self.num_partitions)

        # get access token first so we can reuse it
        if auth_client_id:
            auth_access_token = fhir_get_access_token(
                logger=self.logger,
                server_url=server_url,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                log_level=log_level,
            )

        self.logger.info(
            f"Calling {server_url}/{resource_name} with client_id={auth_client_id} and scopes={auth_scopes}"
        )
        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            # read all the files at path into a dataframe
            path_to_files: str = str(file_path)
            try:
                if delta_lake_table:
                    json_df: DataFrame = df.sql_ctx.read.format("delta").load(
                        path_to_files
                    )
                else:
                    json_df = df.sql_ctx.read.text(
                        path_to_files, pathGlobFilter="*.json", recursiveFileLookup=True
                    )
            except AnalysisException as e:
                if str(e).startswith("Path does not exist:"):
                    if progress_logger:
                        progress_logger.write_to_log(f"Folder {path_to_files} is empty")
                    return df
                raise

            row_count: int = json_df.count()
            if not spark_is_data_frame_empty(df=json_df):
                self.logger.info(
                    f"----- Sending {resource_name} (rows={row_count}) to FHIR server {server_url}"
                    f" with batch_size {batch_size}  -----"
                )
                assert batch_size and batch_size > 0
                desired_partitions: int = math.ceil(row_count / batch_size)
                self.logger.info(
                    f"----- Total Batches for {resource_name}: {desired_partitions}  -----"
                )

                # function that is called for each partition
                def send_partition_to_server(
                    partition_index: int, rows: Iterable[Row]
                ) -> Iterable[List[Dict[str, Any]]]:
                    json_data_list: List[Row] = list(rows)
                    logger = get_logger(__name__)
                    if len(json_data_list) == 0:
                        yield []
                    print(
                        f"Sending batch {partition_index}/{desired_partitions} "
                        f"containing {len(json_data_list)} rows "
                        f"for operation {operation} "
                        f"to {server_url}/{resource_name}. [{name}].."
                    )
                    request_id_list: List[str] = []
                    responses: List[Dict[str, Any]] = []
                    if operation == self.FHIR_OPERATION_DELETE:
                        item: Row
                        # FHIR doesn't support bulk deletes, so we have to send one at a time
                        responses = [
                            send_fhir_delete(
                                obj_id=(
                                    item if "id" in item else json.loads(item["value"])
                                )[
                                    "id"
                                ],  # parse the JSON and extract the id
                                server_url=server_url,
                                resource=resource_name,
                                logger=self.logger,
                                auth_server_url=auth_server_url,
                                auth_client_id=auth_client_id,
                                auth_client_secret=auth_client_secret,
                                auth_login_token=auth_login_token,
                                auth_scopes=auth_scopes,
                                auth_access_token=auth_access_token,
                                log_level=log_level,
                            )
                            for item in json_data_list
                        ]
                    elif operation == self.FHIR_OPERATION_MERGE:
                        if batch_size == 1:
                            # ensure we call one at a time. Partitioning does not guarantee that each
                            #   partition will have exactly the batch size
                            auth_access_token1: Optional[str] = auth_access_token
                            i: int = 0
                            count: int = len(json_data_list)
                            for item in json_data_list:
                                i += 1
                                print(
                                    f"Sending {i} of {count} from partition {partition_index} "
                                    f"for {resource_name}: {json.dumps(item.asDict(recursive=True), default=str)}"
                                )
                                current_bundle: List[Dict[str, Any]]
                                result: Optional[
                                    FhirMergeResponse
                                ] = send_json_bundle_to_fhir(
                                    json_data_list=[
                                        json.dumps(
                                            item.asDict(recursive=True), default=str
                                        )
                                        if "id" in item
                                        else item["value"]
                                    ],
                                    server_url=server_url,
                                    validation_server_url=validation_server_url,
                                    resource=resource_name,
                                    logger=self.logger,
                                    auth_server_url=auth_server_url,
                                    auth_client_id=auth_client_id,
                                    auth_client_secret=auth_client_secret,
                                    auth_login_token=auth_login_token,
                                    auth_scopes=auth_scopes,
                                    auth_access_token=auth_access_token1,
                                    log_level=log_level,
                                    retry_count=retry_count,
                                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                                )
                                if result:
                                    auth_access_token1 = result.access_token
                                    if result.request_id:
                                        request_id_list.append(result.request_id)
                                    responses.extend(result.responses)
                        else:
                            # send a whole batch to the server at once
                            result = send_json_bundle_to_fhir(
                                json_data_list=[
                                    json.dumps(item.asDict(recursive=True))
                                    if "id" in item
                                    else item["value"]
                                    for item in json_data_list
                                ],
                                server_url=server_url,
                                validation_server_url=validation_server_url,
                                resource=resource_name,
                                auth_server_url=auth_server_url,
                                auth_client_id=auth_client_id,
                                auth_client_secret=auth_client_secret,
                                auth_login_token=auth_login_token,
                                auth_scopes=auth_scopes,
                                auth_access_token=auth_access_token,
                                logger=self.logger,
                                log_level=log_level,
                                retry_count=retry_count,
                                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            )
                            if result and result.request_id:
                                request_id_list.append(result.request_id)
                            if result:
                                responses = result.responses
                    # each item in responses is either a json object
                    #   or a list of json objects
                    error_count: int = 0
                    updated_count: int = 0
                    created_count: int = 0
                    deleted_count: int = 0
                    errors: List[str] = []
                    response: Dict[str, Any]
                    for response in responses:
                        if "updated" in response and response["updated"] is True:
                            updated_count += 1
                        if "created" in response and response["created"] is True:
                            created_count += 1
                        if "deleted" in response and response["deleted"] is True:
                            deleted_count += 1
                        if "issue" in response and response["issue"] is not None:
                            error_count += 1
                            errors.append(json.dumps(response["issue"]))
                    print(
                        f"Received response for batch {partition_index}/{desired_partitions} "
                        f"request_ids:[{', '.join(request_id_list)}] "
                        f"total={len(json_data_list)}, error={error_count}, "
                        f"created={created_count}, updated={updated_count}, deleted={deleted_count} "
                        f"to {server_url}/{resource_name}. "
                        f"[{name}] "
                        f"Response={json.dumps(responses, default=str)}"
                    )
                    if progress_logger:
                        progress_logger.log_progress_event(
                            event_name="FhirSender",
                            current=partition_index,
                            total=desired_partitions,
                            event_format_string="{0}: {1}/{2} "
                            + f"{resource_name} "
                            + "completed",
                        )
                    if len(errors) > 0:
                        logger.error(
                            f"---- errors for {partition_index}/{desired_partitions} "
                            f"to {server_url}/{resource_name} -----"
                        )
                        for error in errors:
                            logger.error(error)
                        logger.error("---- end errors -----")

                    yield responses

                # ---- Now process all the results ----
                rdd: RDD[Union[List[Dict[str, Any]], List[List[Dict[str, Any]]]]] = (
                    json_df.repartition(desired_partitions)
                    .rdd.mapPartitionsWithIndex(send_partition_to_server)
                    .cache()
                )

                # turn list of list of string to list of strings
                rdd_type = Union[Dict[str, Any], List[Dict[str, Any]]]
                rdd_flat: RDD[rdd_type] = rdd.flatMap(lambda a: a).filter(lambda x: True)  # type: ignore

                # check if RDD contains a list.  If so, flatMap it
                rdd_first_row_obj = rdd_flat.take(1)
                assert isinstance(rdd_first_row_obj, list), type(rdd_first_row_obj)
                if len(rdd_first_row_obj) > 0:
                    rdd_first_row = rdd_first_row_obj[0]
                    if isinstance(rdd_first_row, list):
                        rdd1: RDD[Dict[str, Any]] = rdd_flat.flatMap(lambda a: a).filter(lambda x: True)  # type: ignore
                    else:
                        rdd1 = rdd_flat  # type: ignore
                    # schema = StructType(
                    #     [
                    #         StructField(name="id", dataType=StringType()),
                    #         StructField(name="created", dataType=BooleanType()),
                    #         StructField(name="updated", dataType=BooleanType()),
                    #         StructField(name="deleted", dataType=BooleanType()),
                    #         StructField(name="resource_version", dataType=StringType()),
                    #         StructField(name="resourceType", dataType=StringType()),
                    #         StructField(name="message", dataType=StringType()),
                    #         StructField(name="issue", dataType=StringType()),
                    #     ]
                    # )
                    try:
                        result_df: DataFrame = rdd1.toDF()
                        self.logger.info(
                            f"Executing requests and writing FHIR {resource_name} responses to disk..."
                        )
                        result_df.write.mode(mode).json(str(response_path))
                        # result_df.show(truncate=False, n=100)
                        self.logger.info(
                            f"Reading from disk and counting rows for {resource_name}..."
                        )
                        result_df = df.sql_ctx.read.json(str(response_path))
                        file_row_count: int = result_df.count()
                        self.logger.info(
                            f"Wrote {file_row_count} FHIR {resource_name} responses to {response_path}"
                        )
                        if view:
                            result_df.createOrReplaceTempView(view)

                        if "issue" in result_df.columns:
                            # if there are any errors then raise exception
                            first_error_response: Optional[Row] = result_df.filter(
                                col("issue").isNotNull()
                            ).first()
                            if first_error_response is not None:
                                if throw_exception_on_validation_failure:
                                    raise FhirSenderValidationException(
                                        url=validation_server_url or server_url,
                                        json_data=json.dumps(
                                            first_error_response[0],
                                            indent=2,
                                            default=str,
                                        ),
                                    )
                                else:
                                    self.logger.info(
                                        f"------- Failed validations for {resource_name} ---------"
                                    )
                                    failed_df = result_df.filter(
                                        col("issue").isNotNull()
                                    )
                                    if error_view:
                                        failed_df.createOrReplaceTempView(error_view)
                                    failed_validations: List[str] = (
                                        failed_df.select("issue")
                                        .rdd.flatMap(lambda x: x)
                                        .collect()
                                    )
                                    self.logger.info(failed_validations)
                                    self.logger.info(
                                        f"------- End Failed validations for {resource_name} ---------"
                                    )
                    except Exception as e:
                        self.logger.exception(f"Exception in FHIR Sender: {str(e)}")
                        self.logger.error(f"Response: {rdd1.collect()}")
                        raise e

        self.logger.info(
            f"----- Finished sending {resource_name} (rows={row_count}) to FHIR server {server_url}  -----"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getServerUrl(self) -> str:
        return self.getOrDefault(self.server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResponsePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.response_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResource(self) -> str:
        return self.getOrDefault(self.resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> Optional[int]:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or f" - {self.getResource()}"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getValidationServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.validation_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getThrowExceptionOnValidationFailure(self) -> Optional[bool]:
        return self.getOrDefault(self.throw_exception_on_validation_failure)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOperation(self) -> str:
        return self.getOrDefault(self.operation)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIgnoreStatusCodes(self) -> Optional[List[int]]:
        return self.getOrDefault(self.ignore_status_codes)



import json
import math
from typing import Any, Dict, Iterable, List, Optional

from pyspark import RDD
from pyspark.ml.param import Param
from pyspark.sql.functions import col
from pyspark.sql.types import Row
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v1.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class HttpDataSender(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        source_view: str,
        view: Optional[str] = None,
        url: Optional[str] = None,
        auth_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        parse_response_as_json: Optional[bool] = True,
        content_type: str = "application/x-www-form-urlencoded",
        post_as_json_formatted_string: Optional[bool] = None,
        batch_count: Optional[int] = None,
        batch_size: Optional[int] = None,
    ):
        """
        Sends data to http server (usually REST API)


        :param source_view: read the payload (body) from this view
        :param view: (Optional) view to put the responses in
        :param url: url to call
        :param auth_url: (Optional) url to use to authenticate with client credentials
        :param client_id: (Optional) client id to use to authenticate with client credentials
        :param client_secret: (Optional) client secret to use to authenticate with client credentials
        :param parse_response_as_json: (Optional) whether to parse response as json or not (default = True)
        :param content_type: content_type to use when posting
        :param batch_count: (Optional) number of batches to create
        :param batch_size: (Optional) max number of items in a batch
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.url: Param[Optional[str]] = Param(self, "url", "")
        self._setDefault(url=None)

        self.auth_url: Param[Optional[str]] = Param(self, "auth_url", "")
        self._setDefault(auth_url=None)

        self.client_id: Param[Optional[str]] = Param(self, "client_id", "")
        self._setDefault(client_id=None)

        self.client_secret: Param[Optional[str]] = Param(self, "client_secret", "")
        self._setDefault(client_secret=None)

        self.content_type: Param[str] = Param(self, "content_type", "")
        self._setDefault(content_type=None)

        self.batch_count: Param[Optional[int]] = Param(self, "batch_count", "")
        self._setDefault(batch_count=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=None)

        self.parse_response_as_json: Param[Optional[bool]] = Param(
            self, "parse_response_as_json", ""
        )
        self._setDefault(parse_response_as_json=None)

        self.post_as_json_formatted_string: Param[Optional[bool]] = Param(
            self, "post_as_json_formatted_string", ""
        )
        self._setDefault(post_as_json_formatted_string=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        name: Optional[str] = self.getName()
        auth_url: Optional[str] = self.getOrDefault(self.auth_url)
        client_id: Optional[str] = self.getOrDefault(self.client_id)
        client_secret: Optional[str] = self.getOrDefault(self.client_secret)
        progress_logger = self.getProgressLogger()
        source_view: str = self.getOrDefault(self.source_view)
        view: Optional[str] = self.getOrDefault(self.view)
        url: Optional[str] = self.getOrDefault(self.url)
        parse_response_as_json: Optional[bool] = self.getOrDefault(
            self.parse_response_as_json
        )
        content_type: str = self.getOrDefault(self.content_type)
        batch_count: Optional[int] = self.getOrDefault(self.batch_count)
        batch_size: Optional[int] = self.getOrDefault(self.batch_size)
        post_as_json_formatted_string: Optional[bool] = self.getOrDefault(
            self.post_as_json_formatted_string
        )

        df = df.sparkSession.table(source_view)

        if spark_is_data_frame_empty(df=df):
            return df

        headers: Dict[str, Any] = {}

        if client_id and auth_url and client_secret:
            # first call auth to get a token
            oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
                OAuth2ClientCredentialsFlow(
                    auth_url=auth_url,
                    client_id=client_id,
                    client_secret=client_secret,
                    progress_logger=progress_logger,
                )
            )

            access_token: Optional[str] = oauth2_client_credentials_flow.get_token()

            if progress_logger:
                progress_logger.write_to_log(
                    f"Received token from {auth_url}: {access_token}"
                )

            if access_token:
                headers = {"Authorization": f"Bearer {access_token}"}

        if progress_logger:
            progress_logger.write_to_log(
                f"Using headers: {json.dumps(headers, default=str)}"
            )

        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            # function that is called for each partition
            # noinspection PyUnusedLocal
            def send_partition_to_server(
                partition_index: int, rows: Iterable[Row]
            ) -> Iterable[Row]:
                json_data_list: List[Dict[str, Any]] = [r.asDict() for r in rows]
                # logger = get_logger(__name__)
                if len(json_data_list) == 0:
                    yield Row(
                        url=None, status=0, result=None, request_type=None, headers=None
                    )

                assert url
                json_data: Dict[str, Any]
                for json_data in json_data_list:
                    headers["Content-Type"] = content_type
                    request: HelixHttpRequest = HelixHttpRequest(
                        request_type=RequestType.POST,
                        url=url,
                        headers=headers,
                        payload=json_data,
                        post_as_json_formatted_string=post_as_json_formatted_string,
                    )
                    if parse_response_as_json:
                        response_json = request.get_result()
                        yield Row(
                            url=url,
                            status=response_json.status,
                            result=response_json.result,
                            headers=json.dumps(headers, default=str),
                            request_type=str(RequestType.POST),
                        )
                    else:
                        response_text = request.get_text()
                        yield Row(
                            url=url,
                            status=response_text.status,
                            result=response_text.result,
                            headers=json.dumps(headers, default=str),
                            request_type=str(RequestType.POST),
                        )

            desired_partitions: int
            if batch_count:
                desired_partitions = batch_count
            else:
                row_count: int = df.count()
                desired_partitions = (
                    math.ceil(row_count / batch_size)
                    if batch_size and batch_size > 0
                    else row_count
                )

            # ---- Now process all the results ----
            rdd: RDD[Row] = (
                df.repartition(desired_partitions)
                .rdd.mapPartitionsWithIndex(send_partition_to_server)
                .cache()
            )

            result_df = rdd.toDF().where(col("url").isNotNull())
            if view:
                result_df.createOrReplaceTempView(view)

            return result_df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def getView(self) -> Optional[str]:
    #     return self.getOrDefault(self.view)  # type: ignore



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v1.fhir_sender import FhirSender
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

import requests


def test_fhir_sender_merge_on_delta(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    test_files_dir: Path = data_dir.joinpath("test_files/patients")

    # convert to delta
    test_files_dir_delta = temp_folder.joinpath("test_files/patients")
    df: DataFrame = spark_session.read.format("json").load(str(test_files_dir))
    df.write.format("delta").save(str(test_files_dir_delta))

    df = spark_session.read.format("delta").load(str(test_files_dir_delta))

    response_files_dir: Path = temp_folder.joinpath("patients-response")

    df = create_empty_dataframe(spark_session=spark_session)

    fhir_server_url: str = "http://fhir:3000/4_0_0"

    # Act
    with ProgressLogger() as progress_logger:
        FhirSender(
            resource="Patient",
            server_url=fhir_server_url,
            file_path=test_files_dir_delta,
            response_path=response_files_dir,
            progress_logger=progress_logger,
            batch_size=1,
            delta_lake_table="table",
        ).transform(df)

    # Assert
    response = requests.get(f"{fhir_server_url}/Patient/00100000000")
    assert response.ok
    json_text: str = response.text
    obj = json.loads(json_text)
    assert obj["birthDate"] == "2017-01-01"

    response = requests.get(f"{fhir_server_url}/Patient/00200000000")
    assert response.ok
    json_text = response.text
    obj = json.loads(json_text)
    assert obj["birthDate"] == "1984-01-01"



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v1.fhir_sender import FhirSender
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

import requests


def test_send_delete(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)
    initial_state_files_dir: Path = data_dir.joinpath("test_files/patients")
    to_delete_files_dir: Path = data_dir.joinpath("test_files/patients_delete")
    initial_state_response_files_dir: Path = temp_folder.joinpath("patients-response")
    delete_response_files_dir: Path = temp_folder.joinpath("patients-response-delete")
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    fhir_server_url: str = "http://fhir:3000/4_0_0"
    with ProgressLogger() as progress_logger:
        FhirSender(
            resource="Patient",
            server_url=fhir_server_url,
            file_path=initial_state_files_dir,
            response_path=initial_state_response_files_dir,
            progress_logger=progress_logger,
        ).transform(df)

    # Act
    with ProgressLogger() as progress_logger:
        FhirSender(
            resource="Patient",
            server_url=fhir_server_url,
            file_path=to_delete_files_dir,
            response_path=delete_response_files_dir,
            progress_logger=progress_logger,
            operation=FhirSender.FHIR_OPERATION_DELETE,
        ).transform(df)

    # Assert
    response = requests.get(f"{fhir_server_url}/Patient/00100000000")
    assert response.status_code == 404

    response = requests.get(f"{fhir_server_url}/Patient/00200000000")
    assert response.ok
    json_text = response.text
    obj = json.loads(json_text)
    assert obj["birthDate"] == "1984-01-01"
